#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "macros.h"
#include "failures.h"
#include "symbol.h"
#include "xmalloc.h"
#include "as_tree.h"

static struct ast_node *ast_node_new(size_t size)
{
	struct ast_node *node;
	
	node = xmalloc0(size);
	node->ref_count = 1;

	return node;	
}

void ast_node_ref(struct ast_node *node)
{
	return_if_fail(node != NULL);

	++node->ref_count;
}

static void ast_node_free(struct ast_node *node)
{
	return_if_fail(node != NULL);

	if (node->finalize != NULL)
		node->finalize(node);
}

void ast_node_unref(struct ast_node *node)
{
	return_if_fail(node != NULL);
	return_if_fail(node->ref_count != 0);

	--node->ref_count;

	if (node->ref_count == 0)
		ast_node_free(node);
}

/* Constructor and destructor of a root node */
static void ast_node_root_free(struct ast_node *node)
{
	return_if_fail(node != NULL);

	if (node->child != NULL)
		ast_node_unref(node->child);

	xfree(node);
}

struct ast_node_root *ast_node_root(struct ast_node *tree)
{
	struct ast_node_root *root;

	return_val_if_fail(tree != NULL, NULL);

	root = (struct ast_node_root *)ast_node_new(sizeof(*root));

	AST_NODE(root)->type = NODE_TYPE_ROOT;
	AST_NODE(root)->child = tree;
	AST_NODE(root)->finalize = ast_node_root_free;

	tree->parent = AST_NODE(root);

	return root;
}

/* Constructor and destructor of a stub node */
static void ast_node_stub_free(struct ast_node *node)
{
	return_if_fail(node != NULL);

	xfree(node);
}

struct ast_node *ast_node_stub()
{
	struct ast_node_stub *node;
	
	node = (struct ast_node_stub *)ast_node_new(sizeof(*node));

	AST_NODE(node)->type = NODE_TYPE_STUB;
	AST_NODE(node)->finalize = ast_node_stub_free;
		
	return AST_NODE(node);
}

/* Constructor and destructor of an id node */
static void ast_node_id_free(struct ast_node *node)
{
	struct ast_node_id *id;

	return_if_fail(node != NULL);

	id = (struct ast_node_id *)node;

	xfree(id->name);
	xfree(node);
}

struct ast_node_id *ast_node_id(data_type_t decl_type, const char *name)
{
	struct ast_node_id *id;
	
	return_val_if_fail(name != NULL, NULL);
	
	id = (struct ast_node_id *)ast_node_new(sizeof(*id));

	id->name = xstrdup(name);
	id->decl_type = decl_type;

	AST_NODE(id)->type = NODE_TYPE_ID;
	AST_NODE(id)->finalize = ast_node_id_free;
	
	return id;
}

struct ast_node_refer *ast_node_refer(data_type_t decl_type, const char *name)
{
	struct ast_node_refer *refer;

	return_val_if_fail(name != NULL, NULL);

	refer = (struct ast_node_refer *)ast_node_new(sizeof(*refer));

	refer->name = xstrdup(name);
	refer->decl_type = decl_type;

	AST_NODE(refer)->type = NODE_TYPE_REFER;
	AST_NODE(refer)->finalize = ast_node_id_free;

	return refer;
}

/* Constructor and destructor of an access node */
static void ast_node_access_free(struct ast_node *node)
{
	struct ast_node_access *access;
	unsigned int i;

	return_if_fail(node != NULL);

	access = (struct ast_node_access *)node;

	xfree(access->name);

	for(i = 0; i < access->nidxs; i++)
		ast_node_unref(access->idxs[i]);

	xfree(access->idxs);

	xfree(node);
}

struct ast_node_access *ast_node_access(data_type_t decl_type, const char *name, unsigned int *dims)
{
	struct ast_node_access *access;
	
	return_val_if_fail(name != NULL, NULL);
	return_val_if_fail(dims != NULL, NULL);
	
	access = (struct ast_node_access *)ast_node_new(sizeof(*access));

	access->decl_type = decl_type;
	access->name = xstrdup(name);
	access->dims = dims;

	AST_NODE(access)->type = NODE_TYPE_ACCESS;
	AST_NODE(access)->finalize = ast_node_access_free;
	
	return access;
}

void ast_node_access_add_idx(struct ast_node_access *access, struct ast_node *idx)
{
	return_if_fail(access != NULL);
	return_if_fail(idx != NULL);

	idx->parent = AST_NODE(access);

	access->nidxs++;
	access->idxs = xrealloc(access->idxs, sizeof(struct ast_node *)*access->nidxs);
	access->idxs[access->nidxs-1] = idx;
}

/* Constructor and destructor of an constant node */
static void ast_node_const_free(struct ast_node *node)
{
	struct ast_node_const *_const;

	return_if_fail(node != NULL);

	_const = (struct ast_node_const *)node;

	if (_const->data_type == DATA_TYPE_STRING)
		xfree(_const->str);

	xfree(node);
}

struct ast_node_const *ast_node_const(data_type_t type, void *data)
{
	struct ast_node_const *_const;
	
	return_val_if_fail(data != NULL, NULL);

	_const = (struct ast_node_const *)ast_node_new(sizeof(*_const));

        switch (type) {
	case DATA_TYPE_INTEGER:
		_const->integer = *(int *)data;
		break;
	case DATA_TYPE_CHAR:
		_const->ch = *(char *)data;
		break;
	case DATA_TYPE_REAL:
		_const->real = *(float *)data;
		break;
	case DATA_TYPE_STRING:
		_const->str = xstrdup(data);
		break;
	default:
		xerror(1, "unexpected data type");
	}
	
	_const->data_type = type;

	AST_NODE(_const)->type = NODE_TYPE_CONST;
	AST_NODE(_const)->finalize = ast_node_const_free;
	
	return _const;
}

/* Constructor and destructor of an operation node */
static void ast_node_op_free(struct ast_node *node)
{
	struct ast_node_op *op;

	return_if_fail(node != NULL);

	op = (struct ast_node_op *)node;

	ast_node_unref(op->left);
	ast_node_unref(op->right);

	xfree(node);
}

struct ast_node_op *ast_node_op(const char *op_helper, struct ast_node *left, struct ast_node *right)
{
	struct ast_node_op *operation;
	
	return_val_if_fail(left != NULL, NULL);
	return_val_if_fail(right != NULL, NULL);
	
	operation = (struct ast_node_op *)ast_node_new(sizeof(*operation));
	
	switch (op_helper[0]) {
	case '+':
		AST_NODE(operation)->type = NODE_TYPE_SUM_OP;
		operation->opcode = OPCODE_ADD;
		break;
	case '-':
		AST_NODE(operation)->type = NODE_TYPE_SUM_OP;
		operation->opcode = OPCODE_SUB;
		break;
	case '*':
		AST_NODE(operation)->type = NODE_TYPE_MUL_OP;
		operation->opcode = OPCODE_MUL;
		break;
	case '/':
		AST_NODE(operation)->type = NODE_TYPE_MUL_OP;
		operation->opcode = OPCODE_DIV;
		break;
	case '%':
		AST_NODE(operation)->type = NODE_TYPE_MUL_OP;
		operation->opcode = OPCODE_MOD;
		break;
	case '>':
		AST_NODE(operation)->type = NODE_TYPE_REL_OP;
		if (op_helper[1] == '=')
			operation->opcode = OPCODE_GE;
		else
			operation->opcode = OPCODE_GT;
		break;
	case '<':
		AST_NODE(operation)->type = NODE_TYPE_REL_OP;
		if (op_helper[1] == '=')
			operation->opcode = OPCODE_LE;
		else
			operation->opcode = OPCODE_LT;
		break;
	case '!':
		AST_NODE(operation)->type = NODE_TYPE_REL_OP;
		if (op_helper[1] == '=')
			operation->opcode = OPCODE_NE;
		else
			xerror(1, "unexpected operation - %s", op_helper);
		break;
	case '=':
		AST_NODE(operation)->type = NODE_TYPE_REL_OP;
		if (op_helper[1] == '=')
			operation->opcode = OPCODE_EQ;
		else
			xerror(1, "unexpected operation - %s", op_helper);
		break;
	case '&':
		AST_NODE(operation)->type = NODE_TYPE_AND_OP;
		if (op_helper[1] == '&')
			operation->opcode = OPCODE_AND;
		else
			xerror(1, "unexpected operation = %s", op_helper);
		break;
	case '|':
		AST_NODE(operation)->type = NODE_TYPE_OR_OP;
		if (op_helper[1] == '|')
			operation->opcode = OPCODE_OR;
		else
			xerror(1, "unexpected operation - %s", op_helper);
		break;
	default:
		xerror(1, "unexpected operation - %s", op_helper);
	}
	
	operation->left = left;
	operation->right = right;
		
	AST_NODE(operation)->finalize = ast_node_op_free;

	AST_NODE(operation)->child = left;

	left->parent = AST_NODE(operation);
	left->next = right;
	
	right->parent = AST_NODE(operation);
	
	return operation;
}

/* Constructor and destructor of a not node */
static void ast_node_not_free(struct ast_node *node)
{
	struct ast_node_not *not_node;

	return_if_fail(node != NULL);

	not_node = (struct ast_node_not *)node;

	ast_node_unref(not_node->expr);

	xfree(node);
}

struct ast_node_not *ast_node_not(struct ast_node *expr)
{
	struct ast_node_not *not_node;
	
	return_val_if_fail(expr != NULL, NULL);
	
	not_node = (struct ast_node_not *)ast_node_new(sizeof(*not_node));
	
	not_node->expr = expr;
	
	AST_NODE(not_node)->type = NODE_TYPE_NOT;
	AST_NODE(not_node)->finalize = ast_node_not_free;

	AST_NODE(not_node)->child = expr;
	
	expr->parent = AST_NODE(not_node);
	
	return not_node;
}

/* Constructor and destructor of an assign node */
static void ast_node_assign_free(struct ast_node *node)
{
	struct ast_node_assign *assign;
	unsigned int i;

	return_if_fail(node != NULL);

	assign = (struct ast_node_assign *)node;

	if (assign->right != NULL)
		ast_node_unref(assign->right);

	if (assign->left != NULL) {
		for (i = 0; i < assign->nleft; i++)
			ast_node_unref(assign->left[i]);
		xfree(assign->left);
	}

	xfree(node);
}

struct ast_node_assign *ast_node_assign()
{
	struct ast_node_assign *assign;
	
	assign = (struct ast_node_assign *)ast_node_new(sizeof(*assign));
	
	AST_NODE(assign)->type = NODE_TYPE_ASSIGN;
	AST_NODE(assign)->finalize = ast_node_assign_free;

	return assign;
}

void ast_node_assign_add_left(struct ast_node_assign *assign, struct ast_node *left)
{
	return_if_fail(assign != NULL);
	return_if_fail(left != NULL);

	AST_NODE(left)->parent = AST_NODE(assign);
	assign->nleft++;
	assign->left = xrealloc(assign->left, sizeof(struct ast_node *)*assign->nleft);
	assign->left[assign->nleft-1] = left;
}

void ast_node_assign_set_right(struct ast_node_assign *assign, struct ast_node *right)
{
	return_if_fail(assign != NULL);
	return_if_fail(right != NULL);

	AST_NODE(right)->parent = AST_NODE(assign);
	assign->right = right;
}

/* Constructor and destructor of a variable initialization node */
static void ast_node_var_init_free(struct ast_node *node)
{
	struct ast_node_var_init *var_init;

	return_if_fail(node != NULL);

	var_init = (struct ast_node_var_init *)node;

	if (AST_INIT(node)->symbol != NULL)
		symbol_unref(AST_INIT(node)->symbol);

	ast_node_unref(var_init->expr);

	xfree(var_init);
}

static void ast_node_arg_init_free(struct ast_node *node)
{
	struct ast_node_var_init *var_init;

	return_if_fail(node != NULL);

	var_init = (struct ast_node_var_init *)node;


	ast_node_unref(var_init->expr);
	xfree(var_init);
}

struct ast_node_var_init *ast_node_var_init(struct ast_node *expr)
{
	struct ast_node_var_init *var_init;

	return_val_if_fail(expr != NULL, NULL);

	var_init = (struct ast_node_var_init *)ast_node_new(sizeof(*var_init));
	var_init->expr = expr;

	AST_NODE(var_init)->type = NODE_TYPE_VAR_INIT;
	AST_NODE(var_init)->finalize = ast_node_var_init_free;
	AST_NODE(var_init)->child = expr;

	expr->parent = AST_NODE(var_init);

	return var_init;
}

struct ast_node_var_init *ast_node_arg_init(struct ast_node *expr)
{
	struct ast_node_var_init *arg_init;

	return_val_if_fail(expr != NULL, NULL);

	arg_init = (struct ast_node_var_init *)ast_node_new(sizeof(*arg_init));
	arg_init->expr = expr;

	AST_NODE(arg_init)->type = NODE_TYPE_ARG_INIT;
	AST_NODE(arg_init)->finalize = ast_node_arg_init_free;
	AST_NODE(arg_init)->child = expr;

	expr->parent = AST_NODE(arg_init);

	return arg_init;
}

/* Constructor and destructor of an array index initialization node */
static void ast_node_array_init_free(struct ast_node *node)
{
	struct ast_node_array_init *array_init;
	unsigned int i;

	return_if_fail(node != NULL);

	array_init = (struct ast_node_array_init *)node;

	if (array_init->initvals != NULL) {
		for (i = 0; i < array_init->ninitvals; i++)
			ast_node_unref(array_init->initvals[i].expr);

		xfree(array_init->initvals);
	}
	symbol_unref(AST_INIT(node)->symbol);

	xfree(array_init);
}

struct ast_node_array_init *ast_node_array_init()
{
	struct ast_node_array_init *array_init;

	array_init = (struct ast_node_array_init *)ast_node_new(sizeof(*array_init));

	AST_NODE(array_init)->type = NODE_TYPE_ARRAY_INIT;
	AST_NODE(array_init)->finalize = ast_node_array_init_free;;

	return array_init;
}

void ast_node_array_add_initval(struct ast_node_array_init *array_init, struct ast_node *expr, unsigned int nidx)
{
	struct array_idx idx;

	return_if_fail(array_init != NULL);
	return_if_fail(expr != NULL);

	array_init->ninitvals++;
	array_init->initvals = xrealloc(array_init->initvals,
				array_init->ninitvals*sizeof(struct array_idx));

	idx.expr = expr;
	idx.nidx = nidx;

	array_init->initvals[array_init->ninitvals-1] = idx;
}

void ast_node_init_set_symbol(struct ast_node *node, struct symbol *symbol)
{
	return_if_fail(node != NULL);
	return_if_fail(symbol != NULL);

	AST_INIT(node)->symbol = symbol;
}

/* Constructor and destructor of a if node */
static void ast_node_if_free(struct ast_node *node)
{
	struct ast_node_if *if_node;

	return_if_fail(node != NULL);

	if_node = (struct ast_node_if *)node;

	ast_node_unref(if_node->expr);
	ast_node_unref(if_node->stmt);

	if (if_node->_else != NULL)
		ast_node_unref(if_node->_else);

	xfree(node);
}

struct ast_node_if *ast_node_if(struct ast_node *expr, struct ast_node *stmt, struct ast_node *_else)
{
	struct ast_node_if *if_node;
	
	return_val_if_fail(expr != NULL, NULL);
	return_val_if_fail(stmt != NULL, NULL);
	
	if_node = (struct ast_node_if *)ast_node_new(sizeof(*if_node));
	
	if_node->expr = expr;
	if_node->stmt = stmt;
	if_node->_else = _else;
	
	AST_NODE(if_node)->type = NODE_TYPE_IF;
	AST_NODE(if_node)->finalize = ast_node_if_free;
	AST_NODE(if_node)->child = expr;
	
	expr->parent = AST_NODE(if_node);
	stmt->parent = AST_NODE(if_node);

	if (_else != NULL)
		_else->parent = AST_NODE(if_node);
	
	return if_node;
}

/* Constructor and destructor of a while node */
static void ast_node_while_free(struct ast_node *node)
{
	struct ast_node_while *while_node;

	return_if_fail(node != NULL);

	while_node = (struct ast_node_while *)node;

	ast_node_unref(while_node->expr);
	ast_node_unref(while_node->stmt);

	xfree(node);
}

struct ast_node_while *ast_node_while(struct ast_node *expr, struct ast_node *stmt)
{
	struct ast_node_while *while_node;

	return_val_if_fail(expr != NULL, NULL);
	return_val_if_fail(stmt != NULL, NULL);

	while_node = (struct ast_node_while *)ast_node_new(sizeof(*while_node));

	while_node->expr = expr;
	while_node->stmt = stmt;

	AST_NODE(while_node)->type = NODE_TYPE_WHILE;
	AST_NODE(while_node)->finalize = ast_node_while_free;
	AST_NODE(while_node)->child = expr;

	expr->parent = AST_NODE(while_node);
	stmt->parent = AST_NODE(while_node);

	return while_node;
}

/* Constructor and destructor of a for node */
static void ast_node_for_free(struct ast_node *node)
{
	struct ast_node_for *for_node;

	return_if_fail(node != NULL);

	for_node = (struct ast_node_for *)node;

	if (for_node->optexpr1 != NULL)
		ast_node_unref(for_node->optexpr1);

	if (for_node->optexpr2 != NULL)
		ast_node_unref(for_node->optexpr2);

	if (for_node->optexpr3 != NULL)
		ast_node_unref(for_node->optexpr3);

	ast_node_unref(for_node->stmt);

	xfree(node);
}

struct ast_node_for *ast_node_for(struct ast_node *optexpr1, struct ast_node *optexpr2, struct ast_node *optexpr3, struct ast_node *stmt)
{
	struct ast_node_for *for_node;
	
	return_val_if_fail(stmt != NULL, NULL);

	for_node = (struct ast_node_for *)ast_node_new(sizeof(*for_node));

	AST_NODE(for_node)->type = NODE_TYPE_FOR;
	AST_NODE(for_node)->finalize = ast_node_for_free;

	if (optexpr1 != NULL) {
		AST_NODE(for_node)->child = optexpr1;
		AST_NODE(optexpr1)->parent = AST_NODE(for_node);
	}
	
	if (optexpr2 != NULL)
		AST_NODE(optexpr2)->parent = AST_NODE(for_node);
	
	if (optexpr3 != NULL)
		AST_NODE(optexpr3)->parent = AST_NODE(for_node);
	
	stmt->parent = AST_NODE(for_node);

	for_node->optexpr1 = optexpr1;
	for_node->optexpr2 = optexpr2;
	for_node->optexpr3 = optexpr3;
	for_node->optexpr3 = optexpr3;
	for_node->stmt = stmt;
	
	return for_node;
}

/* Constructor and destructor of a statment node */
static void ast_node_stmt_free(struct ast_node *node)
{
	struct ast_node_stmt *stmt_node;

	return_if_fail(node != NULL);

	stmt_node = (struct ast_node_stmt *)node;

	if (AST_NODE(stmt_node)->child != NULL)
		ast_node_unref(AST_NODE(stmt_node)->child);

	if (AST_NODE(stmt_node)->next != NULL)
		ast_node_unref(AST_NODE(stmt_node)->next);

	xfree(node);
}

struct ast_node *ast_node_stmt(struct ast_node *node)
{
	struct ast_node_stmt *stmt_node;
	
	return_val_if_fail(node != NULL, NULL);

	stmt_node = (struct ast_node_stmt *)ast_node_new(sizeof(*stmt_node));

	AST_NODE(stmt_node)->type = NODE_TYPE_STMT;
	AST_NODE(stmt_node)->finalize = ast_node_stmt_free;
	AST_NODE(stmt_node)->child = node;

	node->parent = AST_NODE(stmt_node);

	return AST_NODE(stmt_node);
}

/* Constructor and destructor of a call function node */
static void ast_node_func_call_free(struct ast_node *node)
{
	struct ast_node_func_call *call;
	unsigned int i;

	return_if_fail(node != NULL);

	call = (struct ast_node_func_call *)node;

	xfree(call->name);

	if (call->args != NULL) {
		for (i = 0; i < call->nargs; i++)
			ast_node_unref(call->args[i]);
		xfree(call->args);
	}

	xfree(node);
}

struct ast_node_func_call *ast_node_func_call(const char *name, data_type_t ret_type)
{
	struct ast_node_func_call *call;
	
	return_val_if_fail(name != NULL, NULL);
	
	call = (struct ast_node_func_call *)ast_node_new(sizeof(*call));

	AST_NODE(call)->type = NODE_TYPE_CALL;
	AST_NODE(call)->finalize = ast_node_func_call_free;
	
	call->name = xstrdup(name);
	call->ret_type = ret_type;
	
	return call;
}

void ast_node_func_call_add_arg(struct ast_node_func_call *func_call, struct ast_node *arg)
{
	return_if_fail(func_call != NULL);

	arg->parent = AST_NODE(func_call);

	func_call->nargs++;
	func_call->args = xrealloc(func_call->args, sizeof(struct ast_node *)*func_call->nargs);
	func_call->args[func_call->nargs-1] = arg;
}

/* Constructor and destructor of a return node */
static void ast_node_return_free(struct ast_node *node)
{
	struct ast_node_return *return_node;

	return_if_fail(node != NULL);

	return_node = (struct ast_node_return *)node;

	if (return_node->ret_val_node != NULL)
		ast_node_unref(return_node->ret_val_node);

	xfree(node);
}

struct ast_node_return *ast_node_return(struct ast_node *ret_val_node, data_type_t ret_type)
{
	struct ast_node_return *ret_node;
	
	ret_node = (struct ast_node_return *)ast_node_new(sizeof(*ret_node));

	ret_node->ret_val_node = ret_val_node;
	ret_node->ret_type = ret_type;

	AST_NODE(ret_node)->type = NODE_TYPE_RETURN;
	AST_NODE(ret_node)->finalize = ast_node_return_free;
	AST_NODE(ret_node)->child = ret_val_node;

	ret_val_node->parent = AST_NODE(ret_node);
		
	return ret_node;
}

/* Constructor and destructor of a function node */
static void ast_node_function_free(struct ast_node *node)
{
	struct ast_node_func *func;

	return_if_fail(node != NULL);

	func = (struct ast_node_func *)node;

	xfree(func->name);
	ast_node_unref(func->func_tree);

	xfree(node);
}

struct ast_node_func *ast_node_function(const char *name, data_type_t ret_type, struct ast_node *tree)
{
	struct ast_node_func *func;
	
	return_val_if_fail(name != NULL, NULL);

	func = (struct ast_node_func *)ast_node_new(sizeof(*func));

	AST_NODE(func)->type = NODE_TYPE_FUNCTION;
	AST_NODE(func)->finalize = ast_node_function_free;
	AST_NODE(func)->child = tree;

	func->name = xstrdup(name);
	func->func_tree = tree;
	func->ret_type = ret_type;

	tree->parent = AST_NODE(func);

	return func;
}

/* Constructor and destructor of a break node */
static void ast_node_break_free(struct ast_node *node)
{
	return_if_fail(node != NULL);
	xfree(node);
}

struct ast_node_break *ast_node_break()
{
	struct ast_node_break *break_node;
	
	break_node = (struct ast_node_break *)ast_node_new(sizeof(*break_node));
	AST_NODE(break_node)->type = NODE_TYPE_BREAK;
	AST_NODE(break_node)->finalize = ast_node_break_free;
	
	return break_node;
}

/* Constructor and destructor of a continue node */
static void ast_node_continue_free(struct ast_node *node)
{
	return_if_fail(node != NULL);
	xfree(node);
}

struct ast_node_continue *ast_node_continue()
{
	struct ast_node_continue *continue_node;
	
	continue_node = (struct ast_node_continue *)ast_node_new(sizeof(*continue_node));
	AST_NODE(continue_node)->type = NODE_TYPE_CONTINUE;
	AST_NODE(continue_node)->finalize = ast_node_continue_free;
	
	return continue_node;
}

/* Constructor and destructor of a scope node */
static void ast_node_scope_free(struct ast_node *node)
{
	struct ast_node_scope *scope_node;

	return_if_fail(node != NULL);

	scope_node = (struct ast_node_scope *)node;
	if (scope_node->stmt != NULL)
		ast_node_unref(scope_node->stmt);

	symbol_table_unref(scope_node->table);

	xfree(node);
}

struct ast_node_scope *ast_node_scope(struct symbol_table *table, struct ast_node *stmt)
{
	struct ast_node_scope *scope;

	return_val_if_fail(table != NULL, NULL);
	return_val_if_fail(stmt != NULL, NULL);

	scope = (struct ast_node_scope *)ast_node_new(sizeof(*scope));

	scope->table = table;
	scope->stmt = stmt;

	AST_NODE(scope)->type = NODE_TYPE_SCOPE;
	AST_NODE(scope)->finalize = ast_node_scope_free;
	AST_NODE(scope)->child = stmt;

	stmt->parent = AST_NODE(scope);

	return scope;
}

/* Constructor and destructor of an end scope node */
static void ast_node_scope_end_free(struct ast_node *node)
{
	return_if_fail(node != NULL);
	xfree(node);
}

struct ast_node *ast_node_scope_end()
{
	struct ast_node_scope_end *end;

	end = (struct ast_node_scope_end *)ast_node_new(sizeof(*end));
	AST_NODE(end)->type = NODE_TYPE_SCOPE_END;
	AST_NODE(end)->finalize = ast_node_scope_end_free;

	return AST_NODE(end);
}

/* Constructor and destructor of a memory area node */
#ifdef _MEM_AREA
static void ast_node_mem_area_free(struct ast_node *node)
{
	struct ast_node_mem_area *mem_area;

	return_if_fail(node != NULL);

	mem_area = (struct ast_node_mem_area *)node;

	if (mem_area->idx != NULL)
		ast_node_unref(mem_area->idx);

	xfree(node);
}

struct ast_node_mem_area *ast_node_mem_area(char mem_type, unsigned int offset,
					    char data_type, int is_unisgn)
{
	struct ast_node_mem_area *mem_area;

	mem_area = (struct ast_node_mem_area *)ast_node_new(sizeof(*mem_area));

	AST_NODE(mem_area)->type = NODE_TYPE_MEM_AREA;
	AST_NODE(mem_area)->finalize = ast_node_mem_area_free;

	switch (mem_type) {
	case 'I':
	case 'Q':
		mem_area->area = mem_type;
		break;
	default:
		xerror(1, "unknown memory area type");
	}

	switch (data_type) {
	case 'B':
	case 'S':
	case 'L':
	case 'F':
		mem_area->mem_data_type = data_type;
		break;
	default:
		xerror(1, "unknown memory area data type");
	}

	mem_area->offset = offset;
	mem_area->nbits = -1;
	mem_area->is_unsign = is_unisgn;

	return mem_area;
}

void ast_node_mem_area_add_idx(struct ast_node_mem_area *mem_area, struct ast_node *idx)
{
	return_if_fail(mem_area != NULL);

	mem_area->idx = idx;
}

void ast_node_mem_area_add_nbits(struct ast_node_mem_area *mem_area, int nbits)
{
	return_if_fail(mem_area != NULL);

	mem_area->nbits = nbits;
}
#endif
